#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "prime.h"
#include "ompadm.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

int ahr_check (prime_table_t tab, long k0, long m, long i, long checks, int threads)
{
	int quit = 0;
	
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		int sts = 1;
		register long endn, n;
		
		endn = m+checks;
		for ( n = m+1+tid ; sts && n <= endn ; n += threads ) {
			sts = ahr_admissible_mod_pi (tab, k0, m, i, n);
			#pragma omp critical(ahr_check)
			{ if ( ! sts ) quit = 1; if ( quit ) sts = 0; }
		}
	}
	return !quit;
}

// determines the optimal i in [i, k0/2-1] that minimizes the diameter of an aHR-sequence for a given k0 and m, does NOT check admissibility
long ahr_best_i_for_m (prime_table_t tab, long k0, long m, long i)
{
	unsigned char *gd, *gu;
	register long mini;
	register int delta, mindelta;
	
	prime_table_nth_gap (tab, &gd, m+k0/2-1-i);
	prime_table_nth_gap (tab, &gu, m+(k0+1)/2-1+i);
	mindelta = delta = 0;  mini = i;
	for ( i++ ; i < k0/2-1 ; i++ ) {
		delta += *gu++ - *(--gd);
		if ( delta <= mindelta ) { mindelta = delta; mini = i; }
	}
	return mini;
}

static inline long ahr_diameter (prime_table_t tab, long k0, long m, long i)
	{ return prime_table_nth_prime (tab, m+k0/2-1-i) + prime_table_nth_prime (tab, m+(k0+1)/2-1+i); }
	
long ahr_least_m_for_i (prime_table_t tab, long k0, long m, long i, long threads)
{
	long maxbad = 0;
	
	#pragma omp parallel num_threads(threads)
	{
		int sts, quit = 0;
		long n;
		
		for ( n = m-1-omp_get_thread_num() ; !quit && n > 0 ; n -= threads ) {
			sts = ahr_admissible_mod_firstp (tab, k0, n, i);
			#pragma omp critical(mainloop)
			{
				if ( ! sts && n > maxbad ) maxbad = n;
				if ( n <= maxbad ) quit = 1;
			}
		}
	}
	m = maxbad+1;
	return m;
}


int main (int argc, char *argv[])
{
	prime_table_t tab;
	double begin;
	int threads;
	long k0, m, m0, i, d, checks;
	
	if ( argc < 2 ) { puts ("ahrsearch k0 [threads]"); return 0; }
	k0 = atol(argv[1]);
	assert ( k0 >= 20 );
	if ( argc > 2 ) threads = atoi(argv[2]); else threads = omp_get_max_threads();
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);

	begin = omp_get_wtime();
	i = prime_pi (k0);
	printf ("pi(%ld) = %ld (%0.1fs)\n", k0, i, omp_get_wtime()-begin);
	prime_table_init (tab, 2, i+k0);
	printf ("Built prime table from p_%ld=%ld to p_%ld=%ld (%.1fs)\n", tab->si, tab->sp, tab->ti, tab->tp, omp_get_wtime()-begin);

	checks = threads *(sqrt(k0)/threads+1);
	
	m = hr_best_m (tab, k0, begin, threads);  i = 0;
	printf ("HR best m=%ld yields d=%ld\n", m, ahr_diameter (tab, k0, m, i));
	i = ahr_best_i_for_m (tab, k0, m, i);  d = ahr_diameter (tab, k0, m, i);
	printf ("Best i=%ld for m=%ld yields d=%ld\n", i, m, d);
	if ( ! ahr_check (tab, k0, m, i, checks, threads) ) { printf ("aHR sequence for m=%ld, i=%ld is inadmissible\n", m, i); return 0; }
	
	for (;;) {
		m0 = m;
		m = ahr_least_m_for_i (tab, k0, m, i, threads);
		if ( m == m0 ) break;
		d = ahr_diameter (tab, k0, m, i);
		printf ("Least m=%ld for i=%ld yields d=%ld\n", m, i, d);
		if ( ! ahr_check (tab, k0, m, i, checks, threads) ) { printf ("aHR sequence for m=%ld, i=%ld is inadmissible\n", m, i); return 0; }
		i = ahr_best_i_for_m (tab, k0, m, i);  d = ahr_diameter (tab, k0, m, i);
		printf ("Best i=%ld for m=%ld yields d=%ld\n", i, m, d);
		if ( ! ahr_check (tab, k0, m, i, checks, threads) ) { printf ("aHR sequence for m=%ld, i=%ld is inadmissible\n", m, i); return 0; }
	}
	printf ("Asymmetric Hensley-Richard sequence for -p_{m+k0/2-1-i},...,-p_{m+1},...,-1,1,p_{m+1},...,p_{m+(k0+1)/2-1+i with m=%ld and i=%ld has diameter d=%ld\n", m, i, d);
	printf ("This result is heuristic, you now need to run\n");
	printf ("    ahrverify %ld %ld %ld\n", k0, m, i);
	printf ("Total elapsed time %.1fs\n", omp_get_wtime()-begin);
	prime_table_clear (tab);
	prime_cleanup();
}
