#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "prime.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

#define MAXB		7
#define L			6	// use 2^L bit words in bitmask array of 2^B words, where B varies dynamicall from 0 to MAXB

// determines whether the sequence -p_{m+k0/2-1-i},-p_{m+1},-1,1,p_{m+1}, ..., p_{m+(k0+1)/2-1+i} is admissible modulo p
// pB points to an integer between 0 and MAXB that is used to optimize the test as the density mod p changes
// requires tab to start at p_{m+1} (this is verified)
int ahr_admissible_mod_p (prime_table_t tab, long k0, long m, long i, long p, int *pB)
{
	uint64_t b[1<<MAXB];
	unsigned char *gap;
	unsigned char *egapu, *egapd;
	register long j, r, x, lastpu, lastpd;
	register int n, B;
	
	lastpu = prime_table_nth_gap (tab, &egapu, m+(k0+1)/2-1+i);
	assert ( tab->si == m+1 && lastpu );
	if ( p < tab->sp || p > lastpu ) return 1;
	lastpd = prime_table_nth_gap (tab, &egapd, m+k0/2-1-i);
	assert ( lastpd );

	B = *pB;
	// Note that we do not need to consider -1 in the main loop, since it only checks residues 0,...,n-1 mod p, where n is the largest multiple of 2^(L+B) less than p, so we always have n-1 < p-1 (since p is odd and n is even)
	for ( j = 0 ;  j < (p>>(L+B)) ; j++ ) {
		memset (b, 0, (1<<B)*sizeof(*b));
		if ( ! j ) b[0] = 1UL<<1;											// we know 1 is in the sequence
		for ( x = (j<<(L+B)) ; x <= lastpu ; x += p ) {
			for ( r = prime_table_next_gap (tab, &gap, x)-x ; r < (1L<<(L+B)) && gap <= egapu ; r+=2*(*gap++) )
				b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			for ( n = 0; n < (1<<B) ; n++ ) if ( ~b[n] ) break;
			if ( n == (1<<B) ) break;
		}
		if ( x <= lastpu ) continue;
		for ( x = p-((j+1)<<(L+B))+1 ; x <= lastpd ; x += p ) {
			for ( r = (1L<<(L+B))-1 + x-prime_table_next_gap (tab, &gap, x) ; r >= 0 && gap <= egapd ; r-=2*(*gap++) )
				b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			for ( n = 0; n < (1<<B) ; n++ ) if ( ~b[n] ) break;
			if ( n== (1<<B) ) break;
		}
		if ( x > lastpd ) { if ( !j && B ) (*pB)--;  if ( j && B < MAXB ) (*pB)++; return 1; }
	}
	if ( B < MAXB ) (*pB)++;
	for ( j <<= L+B ; j < p-1 ; j++ ) {	// we know -1 is in the sequence
		if ( j == 1 ) continue;		// we know 1 is in the squence
		for ( x = j ; x <= lastpu ; x += p ) if ( prime_table_next_gap (tab, &gap, x) == x ) break;
		if ( x > lastpu ) {
			for ( x = p-j ; x <= lastpd ; x += p ) if ( prime_table_next_gap (tab, &gap, x) == x ) break;
			if ( x > lastpd ) return 1;
		}
	}
	return 0;
}


int main (int argc, char *argv[])
{
	prime_table_t tab;
	double begin;
	int threads;
	register long k0, m, i;

	if ( argc < 4 ) { puts ("ahrverify k0 m i [threads]\n    verifies that the asymmetric Hensley-Richards sequence -p_{m+k0/2-1-i},...,-p_{m+1},...,-1,1,...,p_{m+1},p_{m+(k_0+1)/2-1+i} is admissible."); return 0; }
	k0 = atol(argv[1]);  assert ( k0 > 5 );
	m = atol(argv[2]);  assert ( m >= 1 );
	i = atol(argv[3]);  assert ( i >= 0 && i < k0/2-1);
	if ( argc > 4 ) threads = atoi(argv[4]); else threads = omp_get_max_threads();
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);
	
	begin = omp_get_wtime();
	prime_table_init (tab, m+1, m+(k0+1)/2-1+i);
	printf ("Built prime table from p_%ld=%ld to p_%ld=%ld (%.1fs)\n", tab->si, tab->sp, tab->ti, tab->tp, omp_get_wtime()-begin);

	// Note that we only need to check admissible for primes in [p_{m+1},k0] since no element of p_{m+1},...,p_{m+k0} is zero mod p for p < p_{m+1} and admissibility mod p > k0 is automatic.
	#pragma omp parallel num_threads(threads)
	{
		double start;
		unsigned char *pgap;
		int tid, B;
		register long p;

		tid = omp_get_thread_num();
		start = omp_get_wtime();
		p = tab->sp;  pgap = tab->gaps;
		for ( int  i = 0 ; p <= k0 && i < tid ; i++ ) p += 2*(*pgap++);
		B = MAXB;
		while ( p <= k0 ) {
			if ( ! ahr_admissible_mod_p (tab, k0, m, i, p, &B) ) { printf ("Inadmissible modulo p_%ld=%ld, terminating (tid=%d).\n", prime_table_pi (tab, p), p, tid); exit(-1); }
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("verified to p = %ld (%.1fs)\n", p, end-begin); start = end; }
			}
			for ( int i = 0 ; p <= k0 && i < threads ; i++ ) p += 2*(*pgap++);
		}
	}
	printf ("verified to p = %ld (%.1fs)\n", prime_table_nth_prime (tab, prime_table_pi (tab, k0)), omp_get_wtime()-begin);
	printf ("aHR sequence -p_{m+k0/2-1-i},...,-p_{m+1},...,-1,1,...,p_{m+1},...,p_{m+(k_0+1)/2-1+i} for k0=%ld, m=%ld, i=%ld with diameter %ld is admissible.\n", k0, m, i,
		   prime_table_nth_prime(tab, m+k0/2-1-i)+prime_table_nth_prime(tab, m+(k0+1)/2-1+i));
	prime_table_clear (tab);
	prime_cleanup();
	return 0;
}
