#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "bitmap.h"
#include "prime.h"
#include "ompadm.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

int lcmp (const void *p1, const void *p2) { return *(long*)p1 - *(long*)p2; }

static inline long residue_map (long *rm, long k0, long s, long *pd, long m0, long *nlist, long *rlist, long m, long checks, long maxm, prime_table_t ptab, double begin, long threads)
{
	gap_sequence_t seq;
	long n;
	
	build_gap_sequence_k0 (seq, k0, s, pd, m0, nlist, rlist, ptab, begin, threads);
	n = gap_sequence_residue_map (rm, seq, ptab, m, checks,  maxm, threads);
	clear_gap_sequence (seq);
	return n;
}


int main (int argc, char *argv[])
{
	prime_table_t ptab;
	gap_sequence_t seq;
	double begin, start;
	int threads;
	long i, n, k0, d, m, m0, m1, s, t, checks, maxm;
	long *nlist, *rlist, nlistlen, oldlistlen, nlistsize, *rm;
	char filename[256];
	FILE *fp;
	int good;
	
	if ( argc < 2 ) { puts ("greedysearch k0 [s m threads]"); return 0; }
	k0 = atol(argv[1]);
	assert (k0 > 512);
	if ( argc > 2 ) s = atol (argv[2]); else s = 0;
	if ( argc > 3 ) m0 = atol (argv[3]); else m0 = 0;
	if ( argc > 4 ) threads = atoi(argv[4]); else threads = 0;
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);

	if ( ! s ) s = -k0*log(k0)/2+k0/2;						// heuristically good s
	if ( (s&1) ) s++;									// make s even (we sieve all odd integers anyway)
	d = (((long) (k0*log(k0)+k0)>>6)+1)<<6;				// make d a multiple of 64
	printf ("Initial diameter d=%ld\n", d);

	begin = omp_get_wtime();
	maxm = prime_pi (k0);
	prime_table_init (ptab, 2, maxm+1);
	checks = threads * (0.5*sqrt(k0)/threads+1);
	
	if ( ! m0 ) {
		m0 = 2*sqrt(k0);								// heuristically good m0
		if ( m0 > k0/(log(k0)*log(k0)) ) m0 = k0/(log(k0)*log(k0));
	}
	if ( m0 < 97 ) m0 = 97;								// we want  m0 to be big enough so that p_{m0+1} > 512
	printf ("m0=%ld, p0=%ld (%.1fs)\n", m0, prime_table_nth_prime (ptab, m0), omp_get_wtime()-begin);
	nlistsize = k0/(log(k0)*log(k0));
	nlist = malloc (nlistsize*sizeof(*nlist));  rlist = malloc (nlistsize*sizeof(*nlist));
	nlistlen=0; nlist[0] = 0;
	rm = malloc (checks*sizeof(*rm));
	
	for ( m = m0, good = 0 ; good < 4  ; m += checks ) {
		if ( m >= maxm ) { m = maxm; break; }
		n = residue_map (rm, k0, s, &d, m0, nlist, rlist, m, checks, maxm, ptab, begin, threads);
		printf ("m=%ld (plus %ld of %ld) admissibility ratio %ld/%ld = %f (%.1fs)\n", m0, nlistlen, m-m0, n, checks, (double)n/checks, omp_get_wtime()-begin);
		if ( n == checks ) { good++; continue; } else good = 0;
		for ( i = 0 ; i < checks ; i++ ) {
			if ( rm[i] >= 0 ) {
				nlist[nlistlen] = m+1+i;  rlist[nlistlen++] = rm[i];
				if ( nlistlen == nlistsize ) { nlistsize *= 2; nlist = realloc (nlist, nlistsize*sizeof(*nlist)); rlist = realloc (rlist, nlistsize*sizeof(*rlist)); }
			}
		}
		nlist[nlistlen] = 0;
	}
	free (rm);
	m1 = m;
	printf ("m1=%ld (%.1fs)\n", m1, omp_get_wtime()-begin);
	for ( i = 0, m = m0 ; i < nlistlen && nlist[i] == m+1 && !rlist[i] ; i++, m++ );
	printf ("increased m0 from %ld to %ld\n", m0, m);
	m0 = m;  n = i;
	nlistlen -= n;
	for ( i = 0 ; i < nlistlen ; i++ ) { nlist[i] = nlist[i+n]; rlist[i] = rlist[i+n]; }
	
	build_gap_sequence_k0 (seq, k0, s, &d, m0, nlist, rlist, ptab, begin, threads);
	printf ("Rechecking admissibility from p_%ld to p_%ld...\n", m0, m1);

	oldlistlen = nlistlen;
	start = omp_get_wtime();
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		long i, j, r, p;
		
		j = 0;
		for ( i = m0+1+tid ; i <= m1 ; i += threads ) {
			while ( j < oldlistlen && nlist[j] < i ) j++;
			if ( j < oldlistlen && nlist[j] == i ) continue;
			if ( gap_sequence_admissible_mod_p (seq, p=prime_table_nth_prime (ptab, i)) ) continue;
			r = gap_sequence_best_residue (seq, p);
			if ( r >= 0 ) {
				#pragma omp critical(schinzelsearch)
				{
					nlist[nlistlen] = i;  rlist[nlistlen++] = r;
					if ( nlistlen == nlistsize ) { nlistsize *= 2; nlist = realloc (nlist, nlistsize*sizeof(*nlist)); rlist = realloc (rlist, nlistsize*sizeof(*rlist)); }
					nlist[nlistlen] = 0;
					printf ("Fixed admissibility modulo p_%ld\n", i);
				}
			}
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("verified to p_%ld (%.1fs)\n", i, end-begin); start = end; }
			}
		}
	}
	printf ("verified to p_%ld after fixing %ld defects (%.1fs)\n", m, nlistlen-oldlistlen, omp_get_wtime()-begin);
	
	// sort lists by prime for convenience (yes this is a silly way to do it but we want to keep this self-contained)
	struct { long n; long r; } *rec = malloc (nlistlen*sizeof(*rec));
	for ( i = 0 ; i < nlistlen ; i++ ) {rec[i].n = nlist[i]; rec[i].r = rlist[i]; }
	qsort (rec, nlistlen, sizeof(*rec), lcmp);
	for ( i = 0 ; i < nlistlen ; i++ ) {nlist[i] = rec[i].n; rlist[i] = rec[i].r; }
	free (rec);
	
	clear_gap_sequence (seq);
	n = build_gap_sequence_k0 (seq, k0, s, &d, m0, nlist, rlist, ptab, begin, threads);
	printf ("final sieve of [%ld,%ld] yields %ld survivor with diameter d=%ld\n", s, s+d, n, d);
	d = gap_sequence_best_subsequence (&s, &t, seq, k0);
	printf ("best k0-tuple [%ld,%ld] has diameter d=%ld\n", s, t, d);
	clear_gap_sequence (seq);
	n = build_gap_sequence (seq, s, d+1, m0, nlist, rlist, ptab, threads);
	clear_gap_sequence (seq);
	assert (n==k0);
	
	sprintf (filename, "greedy_%ld_%ld.txt", k0, d);
	fp = fopen (filename, "w");
	fprintf (fp, "k0=%ld\n", k0);
	fprintf (fp, "s=%ld\n", s);
	fprintf (fp, "d=%ld\n", d);
	fprintf (fp, "m=%ld\n", m0);
	for ( m1 = m0, i = 0 ; i < nlistlen ; i++ ) { fprintf (fp, "%ld,%ld\n", nlist[i], rlist[i]);  if ( nlist[i] > m1 ) m1 = nlist[i]; }
	fclose (fp);
	free (nlist); free (rlist);

	printf ("Completed greedy sieve of [%ld,%ld] with k0=%ld using the Schinzel bound p_%ld = %ld and greedily sieving at %ld of the next %ld primes, obtaining diameter d=%ld\n",
		   s, t, k0, m0, prime_table_nth_prime (ptab, m0), nlistlen, m1-m0, d);
	printf ("This results is only heuristic, you now need to run\n    greedyverify %s\n", filename);
	prime_table_clear (ptab);
	prime_cleanup();

	printf ("Total time %.1fs\n", omp_get_wtime()-begin);

}
