#ifndef _OMPADM_INCLUDE_
#define _OMPADM_INCLUDE_

#include <stdint.h>
#include "prime.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

static inline long mod (long a, long m) { register long b = a%m;   if ( b < 0 ) b += m; return b; }

// checks whether -p_{m+k0/2-1-i},...,-p_{m+1},...,-1,1,...,p_{m+1},...,p_{m+(k0+1)/2-1+i} is admissible modulo p_n
int ahr_admissible_mod_pi (prime_table_t tab, long k0, long m, long i, long n);

// checks whether -p_{m+k0/2-1},...,-p_{m+1},...,-1,1,...,p_{m+1},...,p_{m+(k0+1)/2-1} is admissible modulo p_{m+1}
static inline int hr_admissible_mod_firstp (prime_table_t tab, long k0, long m)
	{ return ahr_admissible_mod_pi (tab, k0, m, 0, m+1); }

// checks whether -p_{m+k0/2-1-i},...,-p_{m+1},...,-1,1,...,p_{m+1},...,p_{m+(k0+1)/2-1+i} is admissible modulo p_{m+1}
static inline int ahr_admissible_mod_firstp (prime_table_t tab, long k0, long m, long i)
	{ return ahr_admissible_mod_pi (tab, k0, m, i, m+1); }

// heuristically determines least m such that pi_{m+1},...,pi_{m+k0} is admissible
long zhang_best_m (prime_table_t tab, long k0, double begin, int threads);

// heuristically determines least m such that -p_{m+k0/2-1},...,-p_{m+1},...,-1,1,...,p_{m+1},...,p_{m+(k0+1)/2-1} is admissible
long hr_best_m (prime_table_t tab, long k0, double begin, int threads);

// heuristically determines least m such that sieving [s,s+d] at 1 mod 2 and 0 mod p for odd primes p <= p_m yields and admissible k0-tuple
// increases d as required in order to achieve this
long schinzel_best_m (prime_table_t tab, long k0, long s, long *pd, double begin, int threads);	

// sieves [s,s+d-1] at 1 mod 2 and 0 mod p_n for n <= m and builds gap table of the survivors
// x_list points to an optional zero-terminated list of n > m for which the interval will be sieved at r_n mod p_n (ar at 0 mod p_n if rlist is null)
long build_gap_table (gap_table_t gtab, long s, long d, long m, long*xlist, long *rlist, prime_table_t ptab, int threads);

static inline long schinzel_gap_table (gap_table_t gtab, long s, long d, long m, prime_table_t ptab, int threads)
	{ return build_gap_table (gtab, s, d, m, 0, 0 , ptab, threads); }

// increases d as required in order to obtain a sequence that contains an admissible k0-tuple, decreases d if it looks to big, returns number of survivors
long build_gap_table_k0 (gap_table_t gtab, long k0, long s, long *pd, long m,  long *nlist, long *rlist, prime_table_t ptab, double begin, int threads);
	
static inline long schinzel_gap_table_k0 (gap_table_t gtab, long k0, long s, long *pd, long m, prime_table_t ptab, double begin, int threads)
	{ return build_gap_table_k0 (gtab, k0, s, pd, m, 0, 0 , ptab, begin, threads); }

// checks whether sequence defined by gap table is admissible mod p
int gap_table_admissible_mod_p (gap_table_t tab, long p);

// checks admissibility mod p_n for n=m+1 to pi_checks, returns 0 if admissible at all p_n, ow least n for which this is not true
long gap_table_check_admissibility (gap_table_t gtab, prime_table_t ptab, long m, long checks, long maxm, int threads);

// checks inadmissibility mod p_n for n=m+1 to pi_checks, returns 0 if inadmissible at all p_n, ow least n for which this is not true
long gap_table_check_inadmissibility (gap_table_t gtab, prime_table_t ptab, long m, long checks, long maxm, int threads);

// creates a bitmap whose nth bit indicates whether sequence defined by gap table is admissible mod p_{m+1+n}, for n=0 to checks-1
// returns count of admissible primes
long gap_table_admissibility_map (uint64_t *bm, gap_table_t gtab, prime_table_t ptab, long m, long checks, long maxm, int threads);

// return the least minimally occupied residue modulo p, or -1 if sequence is admissible mod p
long gap_table_best_residue (gap_table_t tab, long p);

// sets rm[n] to the least occupied residue class mod p_{m+1+n}, or to -1 if admissible at p_{m+1+n}, for n from 0 to checks-1
// returns count of admissible primes
long gap_table_residue_map (long *rm, gap_table_t gtab, prime_table_t ptab, long m, long checks, long maxm, int threads);


// determines offset of starting point of subsequence of length k0 with minimal diameter
long gap_table_best_start (gap_table_t gtab, long k0);	

// gap sequence encodes a sequence of integers of the same parity as a sequence of gaps
// gaps of length n < 512 are encoded as an 8-bit byte n/2, gaps of length n=a*512+b >= 512 are encoded as a zeros followed by b/2
typedef struct {
	long s, t, k;							// s is the first element of the sequence, t is the last, k is the number of elements in the sequence (one more than the number of gaps)
	unsigned char *gaps, *egap, *firstz, *endz;	// firstz points to the start of the first gap of length >= 512 (egap if none), endz points one past the last gap of length >= 512 (egap if none)
} gap_sequence_t[1];

long build_gap_sequence (gap_sequence_t gseq, long s, long d, long m,  long *xlist, long *rlist, prime_table_t ptab, int threads);
long build_gap_sequence_k0 (gap_sequence_t gseq, long k0, long s, long *pd, long m,  long *nlist, long *rlist, prime_table_t ptab, double begin, int threads);
void clear_gap_sequence (gap_sequence_t gseq);
void dump_gap_sequence (gap_sequence_t seq);

// checks whether sequence defined by seq is admissible mod p
int gap_sequence_admissible_mod_p (gap_sequence_t seq, long p);

long gap_sequence_best_residue (gap_sequence_t seq, long p);

// sets rm[n] to the least occupied residue class mod p_{m+1+n}, or to -1 if admissible at p_{m+1+n}, for n from 0 to checks-1
// retunrs count of admissible primes
long gap_sequence_residue_map (long *rm, gap_sequence_t seq, prime_table_t ptab, long m, long checks, long maxm, int threads);

// determines a minimal diameter subsequence of k0 elements
long gap_sequence_best_subsequence (long *ps, long *pt, gap_sequence_t seq, long k0);

uint64_t *survivor_bitmap (long s, long d, long m, long*nlist, long *rlist, prime_table_t ptab, int threads);
long survivor_bitmap_best_subsequence (long *ps, uint64_t *bm, long n, long k0);

#endif
