#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "bitmap.h"
#include "prime.h"
#include "ompadm.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

int lcmp (const void *p1, const void *p2) { return *(long*)p1 - *(long*)p2; }

static inline long admissibility_map (uint64_t *bm, long k0, long s, long *pd, long m0, long *nlist, long m, long checks, long maxm, prime_table_t ptab, double begin, long threads)
{
	gap_table_t gtab;
	long n;
	
	build_gap_table_k0 (gtab, k0, s, pd, m0, nlist, 0, ptab, begin, threads);
	n = gap_table_admissibility_map (bm, gtab, ptab, m, checks, maxm, threads);
	gap_table_clear (gtab);
	return n;
}

int main (int argc, char *argv[])
{
	prime_table_t ptab;
	gap_table_t gtab;
	double start, begin, r;
	int threads;
	uint64_t *bm;
	long i, j, m, m0, m1, n, k0, d, s, maxm, checks;
	long *nlist, nlistlen, nlistsize, oldlistlen;
	char filename[256];
	FILE *fp;
	int good;
	
	if ( argc < 2 ) { puts ("schinzelsearch2 k0 [s r threads]"); return 0; }
	k0 = atol(argv[1]);
	assert ( k0 > 13 );
	if ( argc > 2 ) s = atol (argv[2]); else s = 0;
	if ( s ) assert ( s > k0/log(k0) );		// don't let s be too small, we want to avoid large gaps
	if ( argc > 3 ) r = atof (argv[3]); else r = 0;
	assert ( r >= 0 && r <= 1 );
	if ( argc > 4 ) threads = atoi(argv[4]); else threads = 0;
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);

	if ( ! s ) s = k0;
	if ( (s&1) ) s++;
	if ( ! r ) r = 0.01;
	checks = threads * (0.5*sqrt(k0)/threads+1);

	begin = omp_get_wtime();
	maxm = prime_pi(k0);
	prime_table_init (ptab, 2, maxm+1);
	d =  (((long)(k0*log(k0)+k0)>>6)+1)<<6;
	m1 = schinzel_best_m (ptab, k0, s, &d, begin, threads);
	n = checks;
	printf ("m1=%ld (%.1fs)\n", m1, omp_get_wtime()-begin);

	m0 = m1/2;
	bm = bm_alloc (checks);
	do {
		m0 = 0.99*m0-checks;
		if ( m0 < 6 ) { m0 = 6; break; }
		n = admissibility_map (bm, k0, s, &d, m0, 0, m0, checks, maxm, ptab, begin, threads);
		printf ("m=%ld, admissibility ratio %ld/%ld = %f vs %f (%.1fs)\n", m0, n, checks, (double)n/checks, r, omp_get_wtime()-begin);
	} while ( (double)n/checks > r );
	
	while ( m0 + checks/2 < m1 ) {
		m = m0 + (m1-m0)/2;
		n = admissibility_map (bm, k0, s, &d, m, 0, m, checks, maxm, ptab, begin, threads);
		printf ("m=%ld, admissibility ratio %ld/%ld = %f vs %f (%.1fs)\n", m, n, checks, (double)n/checks, r, omp_get_wtime()-begin);
		if ( (double)n/checks < r ) m0 = m; else m1 = m;
	}
	while ( (double)n/checks < r ) {
		m0 += checks/2;
		if ( m0 >= maxm ) { m0 = maxm; break; }
		n = admissibility_map (bm, k0, s, &d, m0, 0, m0, checks, maxm, ptab, begin, threads);
		printf ("m=%ld, admissibility ratio %ld/%ld = %f vs %f (%.1fs)\n", m, n, checks, (double)n/checks, r, omp_get_wtime()-begin);
	}
	printf ("m0=%ld (%.1fs)\n", m0, omp_get_wtime()-begin);
	
	nlistsize = m0/2;
	nlist = malloc (nlistsize*sizeof(*nlist));  nlistlen=0; nlist[0] = 0;
	for ( m = m0, good = 0 ; m < maxm ; m += checks ) {
		n = admissibility_map (bm, k0, s, &d, m0, nlist, m, checks, maxm, ptab, begin, threads);
		printf ("m=%ld (plus %ld of %ld) admissibility ratio %ld/%ld = %f vs %f (%.1fs)\n", m0, nlistlen, m-m0, n, checks, (double)n/checks, r, omp_get_wtime()-begin);
		if ( n == checks ) { if ( ++good == 4 ) break; continue; } else good = 0;
		for ( i = bm_min_unset (bm, checks) ; i < checks ; i = bm_next_unset (bm, i+1, checks) ) {
			nlist[nlistlen++] = m+1+i;
			if ( nlistlen == nlistsize ) { nlistsize *= 2; nlist = realloc (nlist, nlistsize*sizeof(*nlist)); printf ("nlist realloc to %ld\n", nlistsize); }
		}
		nlist[nlistlen] = 0;
	}
	if ( m > maxm ) m = maxm;
	m1 = m;
	printf ("m1=%ld (%.1fs)\n", m1, omp_get_wtime()-begin);
	bm_free (bm);
	
	build_gap_table_k0 (gtab, k0, s, &d, m0, nlist, 0, ptab, begin, threads);
	printf ("Rechecking admissibility from p_%ld to p_%ld...\n", m0, m1);

	oldlistlen = nlistlen;
	start = omp_get_wtime();
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		int sts;
		long i, j;
		
		j = 0;
		for ( i = m0+1+tid ; i <= m1 ; i += threads ) {
			while ( j < oldlistlen && nlist[j] < i ) j++;
			if ( j < oldlistlen && nlist[j] == i ) continue;
			sts = gap_table_admissible_mod_p (gtab, prime_table_nth_prime (ptab, i));
			if ( !sts ) {
				#pragma omp critical(schinzelsearch)
				{
					nlist[nlistlen++] = i; 
					if ( nlistlen == nlistsize ) { nlistsize *= 2; nlist = realloc (nlist, nlistsize*sizeof(*nlist));}
					nlist[nlistlen] = 0;
					printf ("Fixed admissibility modulo p_%ld\n", i);
				}
			}
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("verified to p_%ld (%.1fs)\n", i, end-begin); start = end; }
			}
		}
	}
	gap_table_clear (gtab);
	printf ("verified to p_%ld after fixing %ld defects (%.1fs)\n", m, nlistlen-oldlistlen, omp_get_wtime()-begin);
	qsort (nlist, nlistlen, sizeof(*nlist), lcmp);

	build_gap_table_k0 (gtab, k0, s, &d, m0, nlist, 0, ptab, begin, threads);
	n = gap_table_best_start (gtab, k0);
	printf ("final sieve of [%ld,%ld] yields %ld survivor with diameter d=%ld\n", s, s+d, n, d);
	n = gap_table_best_start (gtab, k0);
	s = gap_table_nth_number (gtab, n);
	d = gap_table_nth_number (gtab, n+k0-1) - s;
	printf ("best k0-tuple [%ld,%ld] has diameter d=%ld\n", s, s+d, d);
	gap_table_clear (gtab);
	n = build_gap_table (gtab, s, d+1, m0, nlist, 0, ptab, threads);
	gap_table_clear (gtab);
	assert (n==k0);
	
	sprintf (filename, "schinzel_%ld_%ld.txt", k0, d);
	fp = fopen (filename, "w");
	fprintf (fp, "k0=%ld\n", k0);
	fprintf (fp, "s=%ld\n", s);
	fprintf (fp, "d=%ld\n", d);
	fprintf (fp, "m=%ld\n", m0);
	for ( i = 0 ; i < nlistlen ; i=j ) {
		for ( j = i+1 ; j < nlistlen && nlist[j] == nlist[j-1]+1; j++ );
		if ( nlist[i] < nlist[j-1] )
			fprintf (fp, "%ld-%ld\n", nlist[i], nlist[j-1]);
		else
			fprintf (fp, "%ld\n", nlist[i]);
	}
	fclose (fp);
	
	if ( nlistlen ) m1 = nlist[nlistlen-1]; else m1 = m0;
	printf ("Found Schinzel sequence in [%ld,%ld] with k0=%ld using the prime bound p_%ld = %ld (and %ld of the next %ld primes), obtaining diameter d=%ld\n",
		   s, s+d, k0, m0, prime_table_nth_prime (ptab, m0), nlistlen, m1-m0, d);
	printf ("This results is only heuristic, you now need to run\n    schinzelverify %s\n", filename);
	free (nlist);
	gap_table_clear (gtab);
	prime_table_clear (ptab);
	prime_cleanup();

	printf ("Total time %.1fs\n", omp_get_wtime()-begin);

}
